#import <Cocoa/Cocoa.h>
//#import <Foundation/Foundation.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "main.h"
#include "MainGUI.h"

#define BUFSIZE 4096
#define BUFSIZE2 (1024*128)


NSString *logPath = @"FolderSyncLog.txt";
FILE *logFile;
NSFileManager *fileManager;
NSString *sourcePath = @".";
NSString *destPath = @"./dest";
NSString *eject = @"";
NSDate *currentDate = 0;
NSDictionary *setDateDict = 0;

struct settingDict
{
    char *key;
    NSString **value;
} settingDict[] = 
{
    {"Log", &logPath},
    {"SourceDirectory", &sourcePath},
    {"DestinateDirectory", &destPath},
    {"Eject", &eject},
    {0, 0},
};


int main(int argc, const char *argv[])
{
    return NSApplicationMain(argc, argv);
}


int FolderSync(void)
{

    if (pre())
        return -1;
    
    SyncFolder(sourcePath, destPath);
	post();
    
    return 0;
}

int pre(void)
{
    [currentDate autorelease];
    currentDate = [[NSDate alloc] init];
    [setDateDict autorelease];
	setDateDict = [[NSDictionary dictionaryWithObject: currentDate
                        forKey: NSFileModificationDate] retain];
    if (OpenFiles())
        return -1;
    fileManager = [NSFileManager defaultManager];
    
    [fileManager changeCurrentDirectoryPath: sourcePath];

    return 0;
}


int DoScript(NSString *settingPath)
{
FILE *setting;
char oneLine[BUFSIZE];
int i;
char firstWord[BUFSIZE];

    if (!(setting = fopen([settingPath cString], "r")))
        return -1;

    while (fgets(oneLine, BUFSIZE, setting))
    {
        sscanf(oneLine, "%s", firstWord);
        if (!strcmp(firstWord, "Go"))
        {
            FolderSync();
        }
        for (i = 0; settingDict[i].key; i++)
        {
            if (!strcmp(settingDict[i].key, firstWord))
            {
                if (oneLine[strlen(oneLine)-1] =='\n')
                    oneLine[strlen(oneLine)-1] = 0;
                if (strlen(oneLine) > strlen(settingDict[i].key))
                {
                    settingDict[i].value[0] = [NSString stringWithCString: 
                            oneLine + strlen(settingDict[i].key) + 1];
                }
                else
                {
                    settingDict[i].value[0] = @"";
                }
            }
        }
    }
    
    if (logFile)
        fclose(logFile);
    return 0;
}


int OpenFiles(void)
{
    if (!logFile)
        if (!(logFile = fopen([logPath cString], "w")))
            return -1;
    fprintf(logFile, "%s\n", [[currentDate description] cString]);
    
    
    return 0;
}


void SyncFolder(NSString *sourceDir, NSString *destDir)
{
NSEnumerator *sourceEnumerator = 0;
NSEnumerator *destEnumerator = 0;
NSString *sourceFile;
NSString *destFile;
NSComparisonResult compareResult;
BOOL sourceIsDir;
BOOL destIsDir;
NSArray *sourceArray;
NSArray *destArray;

//fprintf(logFile, "syncFolderFrom: %s\nsyncFolderTo: %s\n",
                //[sourceDir lossyCString], [destDir lossyCString]);
//fflush(logFile);

    sourceArray = [[fileManager directoryContentsAtPath: sourceDir]
                        sortedArrayUsingSelector:@selector(compare:)];
    sourceEnumerator = [sourceArray objectEnumerator];
    destArray = [[fileManager directoryContentsAtPath: destDir]
                        sortedArrayUsingSelector:@selector(compare:)];
    destEnumerator = [destArray objectEnumerator];    
    NextFile(&sourceFile, sourceDir, sourceEnumerator, &sourceIsDir);
    NextFile(&destFile, destDir, destEnumerator, &destIsDir);
    
    while (sourceFile || destFile)
    {        
        if (!destFile)
        {
            compareResult = NSOrderedAscending;
        }
        else if (!sourceFile)
        {
            compareResult = NSOrderedDescending;
        }
        else
        {
            compareResult = [sourceFile compare: destFile];
        }
        
        switch (compareResult)
        {
        case NSOrderedAscending:
        //case NSOrderedDescending:
//fprintf(logFile, "sourceDir: %s\n", [[sourceArray description] lossyCString]);
//fprintf(logFile, "destDir: %s\n", [[destArray description] lossyCString]);
//fprintf(logFile, "comparedWith: %s\n", [destFile lossyCString]);
//fflush(logFile);
            CopyFile([sourceDir stringByAppendingPathComponent: sourceFile], 
                    [destDir stringByAppendingPathComponent: sourceFile]);
            NextFile(&sourceFile, sourceDir, sourceEnumerator, &sourceIsDir);
            break;
        case NSOrderedDescending:
//fprintf(logFile, "sourceDir: %s\n", [[sourceArray description] lossyCString]);
//fprintf(logFile, "destDir: %s\n", [[destArray description] lossyCString]);
//fprintf(logFile, "comparedWith: %s\n", [sourceFile lossyCString]);
//fflush(logFile);
            DeleteFile([destDir stringByAppendingPathComponent: destFile]);
            NextFile(&destFile, destDir, destEnumerator, &destIsDir);
            break;
        case NSOrderedSame:
            SyncFile([sourceDir stringByAppendingPathComponent: sourceFile], 
                    [destDir stringByAppendingPathComponent: destFile],
                    sourceIsDir, destIsDir);
            NextFile(&sourceFile, sourceDir, sourceEnumerator, &sourceIsDir);
            NextFile(&destFile, destDir, destEnumerator, &destIsDir);
            break;
        }
    }
}


void NextFile(NSString **fileP, NSString *directory, NSEnumerator *enumerator, BOOL *isDir)
{
NSDictionary *fileAttr;
NSString *fileType;

    do
    {
        if (!(*fileP = [enumerator nextObject]))
            return;
        fileAttr = [fileManager fileAttributesAtPath:
                [directory stringByAppendingPathComponent: *fileP]
                traverseLink:NO];
        fileType = [fileAttr objectForKey: NSFileType];
        *isDir = [fileType isEqualToString: NSFileTypeDirectory];
    } while (!([fileType isEqualToString: NSFileTypeRegular] || *isDir));
}


void CopyFile(NSString *sourcePath, NSString *destPath)
{
BOOL success;
int ret;

	retry:
fprintf(logFile, "copyFrom: %s\ncopyTo: %s\n", [sourcePath lossyCString],
            [destPath lossyCString]);
    success = [fileManager copyPath: sourcePath toPath: destPath handler: 0];

	if (!success)
	{
		fprintf(logFile, "failed\n");
		fflush(logFile);
		/* Use GUI */
		ret = GUIFailAlert(sourcePath, destPath);
		switch (ret)
		{
		case FailRetry:
			fprintf(logFile, "retry\n");
			goto retry;
			break;
		case FailAbort:
			fprintf(logFile, "abort\n");
			fflush(logFile);		
			exit(0);
			break;
		case FailContinue:
			fprintf(logFile, "continue\n");
			fflush(logFile);		
			break;
		}
	}
}


void DeleteFile(NSString *path)
{
fprintf(logFile, "delete: %s\n", [path lossyCString]);
fflush(logFile);
    [fileManager removeFileAtPath: path handler: 0];
}


void SyncFile(NSString *sourcePath, NSString *destPath,
        BOOL sourceIsDir, BOOL destIsDir)
{
    switch (sourceIsDir + destIsDir)
    {
    case 0:
        SyncRegularFile(sourcePath, destPath);
        break;
    case 1:
        DeleteFile(destPath);
        CopyFile(sourcePath, destPath);
        break;
    case 2:
        SyncFolder(sourcePath, destPath);
        break;
    }
}


void SyncRegularFile(NSString *sourcePath, NSString *destPath)
{
NSDate *sourceDate = [[fileManager fileAttributesAtPath: sourcePath
            traverseLink:NO] objectForKey: NSFileModificationDate];
NSDate *destDate = [[fileManager fileAttributesAtPath: destPath
            traverseLink:NO] objectForKey: NSFileModificationDate];

    if ([sourceDate compare: destDate] == NSOrderedDescending)
    {
fprintf(logFile, "sourceFile is newer\n");
fflush(logFile);
        if ([sourceDate compare: currentDate] == NSOrderedDescending)
        {
            [fileManager changeFileAttributes: setDateDict
                    atPath: sourcePath];
        }
        DeleteFile(destPath);
        CopyFile(sourcePath, destPath);
    }
}

void post()
{
char command[BUFSIZE];
    fprintf(logFile, "Complete.\n");
    if ([eject length])
    {
        [fileManager changeCurrentDirectoryPath: @"/"];
        sprintf(command, "(sleep 2;/usr/sbin/disktool -e %s 0)&",
            [eject fileSystemRepresentation]);
        system(command);
    }
    //fclose(logFile);
}
